require 'open3'

class Foreman::Provision::Ssh
  attr_reader :template, :uuid, :address, :username, :options

  # options: {
  #   template: "/path/to/template",
  #   uuid:     "uuid",
  #   key_data: "private-key-content",
  #   password: "string",
  # }
  def initialize(address, username = "root", options = {})
    @username = username
    @address  = address
    @template = options.delete(:template) || raise("must provide a template")
    @uuid     = options.delete(:uuid)     || "#{address}-#{username}"
    @options  = options
    @tmp_key_path = create_tmp_key

    check_auth_options
    attempt_connection!
  end

  def deploy!
    scp

    logger.debug "about to execute #{main_command}"

    stdout, stderr, status = Open3.capture3(*ssh_cmd(main_command))

    logger.debug(stdout) if stdout.presence
    logger.error(stderr) if stderr.presence

    status.success?
  end

  private

  # Attempt to connect to the remote host and run a simple command
  def attempt_connection!
    start_time = Time.now.to_i

    logger.debug "Attempting to connect to #{remote_host}"
    # TODO: Replace password with ***
    logger.debug ssh_cmd("pwd")

    while (Time.now.to_i - start_time) < Setting[:ssh_timeout].to_i
      _stdout, stderr, status = Open3.capture3(*ssh_cmd("pwd"))

      if status.success?
        logger.info "SSH connection successfully established to #{remote_host}"
        break
      end

      logger.debug stderr.strip
      sleep 2
    end

    raise "Failed to connect to #{address} after #{Setting[:ssh_timeout]} seconds" unless status.success?
  end

  def remote_host
    "#{username}@#{address}"
  end

  def remote_script
    "bootstrap-#{uuid}"
  end

  def command_prefix
    (username == "root") ? "" : "sudo "
  end

  # Use the users home to store the provision script since we can't reliably
  # tell if other locations are writeable or executable by the user.
  def main_command
    main_execution = "(chmod 0701 ./#{remote_script} && #{command_prefix} ./#{remote_script} ; echo $? >#{remote_script}.status) 2>&1"
    cleanup = "#{command_prefix} rm -f #{remote_script}"
    "#{command_prefix} sh -c '#{main_execution} | tee #{remote_script}.log; #{cleanup}; exit $(cat #{remote_script}.status)'"
  end

  def logger
    Rails.logger
  end

  def scp
    logger.debug "about to upload #{template} to remote system at #{remote_script}"
    logger.debug scp_cmd
    _stdout, stderr, status = Open3.capture3(*scp_cmd)

    return if status.success?

    logger.error stderr
    raise Foreman::Exception.new("Failed to upload '#{@template}' to remote system at #{address}")
  end

  def passwd_cmd
    return if @options[:password].empty?
    ["sshpass", "-p", @options[:password]]
  end

  def ssh_key
    return if @options[:key_data].empty?
    ["-i", @tmp_key_path]
  end

  def ssh_cmd(host_command)
    [passwd_cmd, "ssh", ssh_key, ssh_options, remote_host, host_command].flatten.compact
  end

  def scp_cmd
    [passwd_cmd, "scp", ssh_key, ssh_options, @template, "#{remote_host}:~/#{remote_script}"].flatten.compact
  end

  def ssh_options
    opts = ["ConnectTimeout=6", "StrictHostKeyChecking=accept-new",
            "Compression=yes"]

    methods = []
    methods << 'publickey' if @options[:key_data].present?
    methods << 'password' if @options[:password].present?

    opts << "PreferredAuthentications=#{methods.join(',')}"

    opts.map { |o| "-o #{o}" }
  end

  def create_tmp_key
    return if @options[:key_data].empty?

    file = Tempfile.new("ssh-key-#{uuid}")
    file.write(@options[:key_data])
    file.close
    file.path
  end

  def check_auth_options
    return if @options[:key_data].present? || @options[:password].present?

    raise Foreman::Exception.new("Missing :key_data or :password options, no authentication method available")
  end
end
